import os
import pyutilib.th as unittest
from pyutilib.misc.pyyaml_util import *
import coopr.pyomo.scripting.util as util
from coopr.pyomo.base import Var
import coopr.plugins
from coopr.pyomo.base.objective import minimize, maximize
import cPickle
from coopr.pyomo.base.piecewise import Bound, PWRepn

currdir = os.path.dirname(os.path.abspath(__file__))

smoke_problems = ['convex_var']

nightly_problems = ['convex_vararray', 'concave_vararray', \
                'concave_var','piecewise_var', 'piecewise_vararray']

expensive_problems = ['piecewise_multi_vararray', \
                'convex_multi_vararray1','concave_multi_vararray1', \
                'convex_multi_vararray2','concave_multi_vararray2']


def module_available(module):
    try:
        __import__(module)
        return True
    except ImportError:
        return False

def has_gurobi_lp():
    try:
        gurobi = coopr.plugins.solvers.GUROBI(keepFiles=True)
        available = (not gurobi.executable() is None) and gurobi.available(False)
        return available
    except pyutilib.common.ApplicationError:
        return False

def has_gurobi_nl():
    try:
        gurobi = coopr.plugins.solvers.GUROBI(keepFiles=True)
        available = (not gurobi.executable() is None) and gurobi.available(False)
        asl = coopr.plugins.solvers.ASL(keepFiles=True, options={'solver':'gurobi_ampl'})
        return available and (not asl.executable() is None) and asl.available(False)
    except pyutilib.common.ApplicationError:
        return False

def has_gurobi_python():
    if module_available('gurobipy'):
        return True
    return False

def has_cplex_lp():
    try:
        cplex = coopr.plugins.solvers.CPLEX(keepFiles=True)
        available = (not cplex.executable() is None) and cplex.available(False)
        return available
    except pyutilib.common.ApplicationError:
        return False

def has_cplex_nl():
    try:
        cplex = coopr.plugins.solvers.CPLEX(keepFiles=True)
        available = (not cplex.executable() is None) and cplex.available(False)
        asl = coopr.plugins.solvers.ASL(keepFiles=True, options={'solver':'cplexamp'})
        return available and (not asl.executable() is None) and asl.available(False)
    except pyutilib.common.ApplicationError:
        return False

def has_cplex_python():
    if module_available('cplex'):
        return True
    return False

def has_glpk_python():
    if module_available('glpk'):
        return True
    return False

def has_glpk_lp():
    try:
        glpk = coopr.plugins.solvers.GLPK(keepFiles=True)
        available = (not glpk.executable() is None) and glpk.available(False)
        return available
    except pyutilib.common.ApplicationError:
        return False

writer_solver = []
#if has_cplex_python():
#    writer_solver.append(('python','cplex'))
#if has_gurobi_python():
#    writer_solver.append(('python','gurobi'))
if has_cplex_lp():
    writer_solver.append(('lp','cplex'))
#if has_gurobi_lp():
#    writer_solver.append(('lp','gurobi'))
#if has_cplex_nl():
#    writer_solver.append(('nl','cplexamp'))
#if has_gurobi_nl():
#    writer_solver.append(('nl','gurobi_ampl'))
#if has_glpk_lp():
#    writer_solver.append(('lp','glpk'))
#if has_glpk_python():
#    writer_solver.append(('python','glpk'))


def createTestMethod(pName,problem,solver,writer,kwds):
    
    def testMethod(obj):

        from pyutilib.misc import Options
        os.sys.path.append(currdir+'/problems')
        m = __import__(problem)
        os.sys.path.pop()
        options = Options()
        options.solver = solver
        options.solver_io = writer
        options.quiet = True
        options.debug = True
        data = Options(options=options)
        
        model = m.define_model(**kwds)
        instance = model.create()

        opt_data = util.apply_optimizer(data, instance=instance)
        
        instance.load(opt_data.results)
        
        new_results = ( (v[idx].name,v[idx].value) for (n,v) in instance.active_components(Var).iteritems() \
                                                 for idx in v \
                                                 if (n[:2] == 'Fx') \
                                                 or (n[:1] == 'x') )
        baseline_results = getattr(obj,problem+'_results')
        for name, value in new_results:
            if abs(baseline_results[name]-value) > 0.00001:
                raise IOError, "Difference in baseline solution values and current solution values using:\n" + \
                "Solver: "+solver+"\n" + \
                "Writer: "+writer+"\n" + \
                "Variable: "+name+"\n" + \
                "Solution: "+str(value)+"\n" + \
                "Baseline: "+str(baseline_results[name])+"\n"

    return testMethod


def assignTests(cls, problem_list):
    for writer,solver in writer_solver:
        for PROBLEM in problem_list:
            aux_list = ['','force_pw']
            for AUX in aux_list:
                for REPN in PWRepn:
                    for BOUND_TYPE in Bound:
                        for SENSE in [maximize,minimize]:
                            if not(((BOUND_TYPE == Bound.Lower) and (SENSE == maximize)) or \
                                   ((BOUND_TYPE == Bound.Upper) and (SENSE == minimize))):
                                kwds = {}
                                kwds['sense'] = SENSE
                                kwds['pw_repn'] = REPN
                                kwds['pw_constr_type'] = BOUND_TYPE
                                if SENSE == maximize:
                                    attrName = "test_{0}_{1}_{2}_{3}_{4}_{5}".format(PROBLEM,REPN,BOUND_TYPE,'maximize',solver,writer)
                                elif SENSE == minimize:
                                    attrName = "test_{0}_{1}_{2}_{3}_{4}_{5}".format(PROBLEM,REPN,BOUND_TYPE,'minimize',solver,writer)
                                if AUX != '':
                                    kwds[AUX] = True 
                                    attrName += '_'+AUX
                                setattr(cls,attrName,createTestMethod(attrName,PROBLEM,solver,writer,kwds))


@unittest.category('nightly', 'expensive')
@unittest.skipIf(writer_solver==[], "Can't find a solver.")
class PiecewiseLinearTest_Nightly(unittest.TestCase):

    @classmethod
    def setUpClass(self):
        for p in nightly_problems:
            f = open(currdir+'/baselines/'+p+'_baseline_results.pickle','r')
            baseline_results = cPickle.load(f)
            setattr(self,p+'_results',baseline_results)
            f.close()
    
assignTests(PiecewiseLinearTest_Nightly, nightly_problems)


@unittest.category('smoke', 'nightly', 'expensive')
@unittest.skipIf(writer_solver==[], "Can't find a solver.")
class PiecewiseLinearTest_Smoke(unittest.TestCase):

    @classmethod
    def setUpClass(self):
        for p in smoke_problems:
            f = open(currdir+'/baselines/'+p+'_baseline_results.pickle','r')
            baseline_results = cPickle.load(f)
            setattr(self,p+'_results',baseline_results)
            f.close()
        
assignTests(PiecewiseLinearTest_Smoke, smoke_problems)


@unittest.category('expensive')
@unittest.skipIf(writer_solver==[], "Can't find a solver.")
class PiecewiseLinearTest_Expensive(unittest.TestCase):

    @classmethod
    def setUpClass(self):
        for p in expensive_problems:
            f = open(currdir+'/baselines/'+p+'_baseline_results.pickle','r')
            baseline_results = cPickle.load(f)
            setattr(self,p+'_results',baseline_results)
            f.close()
    
assignTests(PiecewiseLinearTest_Expensive, expensive_problems)

if __name__ == "__main__":
    unittest.main()



from coopr.pyomo import *

model = AbstractModel()

model.a = Param(initialize=2)

model.x = Var(within=NonNegativeReals)
model.y = Var(within=NonNegativeReals)
model.z = Var(within=NonNegativeReals,bounds=(7,None))
model.v = Var(within=NonNegativeReals)

def obj_rule(model):
    return 1.0*model.z + 2.0*model.x*model.x + 3.0*model.y + 4.0*model.v**2 - model.x*model.z + 1.5*(model.x+model.z)**2
model.obj = Objective(rule=obj_rule,sense=minimize)

def constr_rule(model):
    return (model.a,model.y*model.y,None)
model.constr = Constraint(rule=constr_rule)

def constr2_rule(model):
    return model.x/model.a >= model.y
model.constr2 = Constraint(rule=constr2_rule)

def constr3_rule(model):
    return model.z <= model.y + model.a
model.constr3 = Constraint(rule=constr3_rule)

def constr4_rule(model):
    return model.v**2 >= 1
model.constr4 = Constraint(rule=constr4_rule)




##
## General description of the API for Pyomo objects
##
## Note: this tutorial was not designed to run as a script
##

#
# Imports
#
import sys
from os.path import abspath,dirname
sys.path.insert(0, dirname(dirname(dirname(dirname(abspath(__file__))))))
currdir = dirname(abspath(__file__))

from coopr.pyomo import *


## -------------------------------------------------------------------------
##
## Model objects
##
## -------------------------------------------------------------------------
#
#  The Model object is a fundamental to the use of Pyomo.  Elements of an
#  application are designed with sets, parameters, and variables; these
#  objects are contained in a model, which supports the generation of
#  problem instances that can be solved.
#
model = AbstractModel()
#
#  Once a model has been setup, a problem instance can be created with the
#  create() method:
#
instance = model.create("foo.dat")


## -------------------------------------------------------------------------
##
## Set objects
##
## -------------------------------------------------------------------------
#
#  The Set object is a container for an unordered set of numeric or
#  symbolic values.
#
####
#### Set Constructor Syntax
####
#
#  Simple constructor of a single set
#
model.A = Set()
#
#  Creating an array of sets, which are indexed with another set
# 
model.C = Set(model.A,model.A)
#
#  Option 'initialize' indicates how values in the set will be constructed.
#  This option behaves differently depending on the type of data provided:
#  (1) a built-in set or list object can be used to initialize a single set,
#  (2) a dictionary can be used to initialize an array of sets, and 
#  (3) a function can be used to initial a set (perhaps using model
#      information.
#
model.A = Set(initialize=[1,4,9])
model.B = Set(model.A, initialize={1:[1,4,9], 2:[2,5,10]})
def f(model):
    return range(0,10)
model.A = Set(initialize=f)
#
#  Option 'ordered' specifies whether the set elements are ordered
#  This option allows for more sophisticated construction rules
#
def f(model, i):
    if i==10:
        return Set.End
    if i==0:
        return 1
    else:
        return model.A[i-1] * (i+1)
model.A = Set(ordered=True, initialize=f)
#
#  Option 'within' specifies a set that is used to validate set elements
#
model.B = Set(within=model.A)
#
#  Option 'validate' specifies a function that is used to validate set elements
#
def f(model, value):
    return value in model.A
model.B = Set(validate=f)
#
#  Option 'dimen' specifies the arity of the data in the set
#
model.A = Set(dimen=3)
#
####
#### Set Methods
####
#
# len() - returns the number of elements of a set
#
len(model.A)
len(model.A[i])
#
# data() - returns the underlying set object used by Set()
#
instance.A.data()
instance.A[i].data()
#
# dim() - returns the number of dimensions of the set index
#
instance.A.dim()
#
# clear() - emptys all sets
#
instance.A.clear()
#
# add() - adds data to a set
#
instance.A.add(1,3,5)
instance.A[i].add(1,3,5)
#
# remove() - removes data from a set, throwing an exception if the data does
#                not exist
#
instance.A.remove(1)
instance.A[i].remove(1)
#
# discard() - removes data from a set without throwing an exception if the
#                data does not exist
#
instance.A.discard(2)
instance.A[i].discard(2)
#
# Set iteration
#
for val in instance.A:
    print val
#
# Set comparisons
#
instance.A < instance.B     # True if A is strict subset of B
instance.A <= instance.B    # True if A is a subset of B
instance.A == instance.B    # True if A equals B
instance.A >= instance.B    # True if A is a superset of B
instance.A > instance.B     # True if A is a strict superset of B
#
# Set membership
#
val in instance.A           # True if 'val' is in A
#
# Set operations
#
instance.A | instance.B     # Set union
instance.A & instance.B     # Set intersection
instance.A ^ instance.B     # Set symmetric difference
instance.A - instance.B     # Set difference
#
# Set cross product - define a new set that is the cross-product of
#       two or more sets
#
instance.A * instance.B
#
# Ordered set operations
#
instance.A[j]               # returns the j'th member of ordered set A
instance.A[i][j]            # returns the j'th member of ordered set A[i]
#
# keys() - returns the indices of the set array
#
instance.A.keys()
#
####
#### Set attributes
####
#
# dimen - Specifies the arity of the data in the set
#
model.A.dimen = 3
#
# virtual - is True if the set contains no concrete data
#
instance.A.virtual


## -------------------------------------------------------------------------
##
## Param objects
##
## -------------------------------------------------------------------------
#
#  The Param object defines a singleton or array of numeric parameters.
#
####
#### Param Constructor Syntax
####
#
#  Singleton parameter
#
model.Z = Param()
#
#  Array of parameters
#
model.Z = Param(model.A,model.B)
#
#  Option 'initialize' specifies values used to construct the parameter
#
model.Z = Param(initialize=9)
model.Z = Param(model.A,initialize={1:1, 2:4, 3:9})
model.Z = Param(model.A,initialize=2)
#
#  Option 'initialize' can also specify a function used to construct the 
#       parameter
#
def f(model, i):
    return 3*i
model.Z = Param(model.A,initialize=f)
#
#  Option 'default' specifies values used for a parameter if no value
#   has been set.  Note that for singleton parameters this has the same
#   role as the 'initialize' option.  For parameter arrays this 
#   'fills in' parameter values that have not been initialized.
#
model.Z = Param(default=9.0)
model.Z = Param(model.A,default=9.0)
#
#  Option 'within' specifies a set that is used to validate parameters
#
model.Z = Param(within=model.A)
#
#  Option 'validate' specifies a function that is used to validate parameters
#
def f(model, value):
    return value in model.A
model.Z = Param(validate=f)
#
####
#### Param Methods
####
#
# len() - returns the number of parameters
#
len(instance.Z)
#
# dim() - returns the number of dimensions of the parameter index
#
instance.Z.dim()
#
# keys() - returns the indices of the parameter array
#
instance.Z.keys()
#
# Coercing parameter values explicitly
#
tmp = float(instance.Z)
tmp = float(instance.Z[i])
#
# Getting parameter values (which may be integer or floats)
#
tmp = value(instance.Z)
tmp = value(instance.Z[i])
tmp = instance.Z.value
tmp = instance.Z[i].value
#
# Setting parameter values
#
instance.Z = tmp
instance.Z[i] = tmp
#
# NOTE: if Z is an array parameter, then
#
#  instance.Z = tmp
#
# will initialize all of the parameter values to the value of 'tmp'.
#


## -------------------------------------------------------------------------
##
## Var objects
##
## -------------------------------------------------------------------------
#
#  The Var object defines a singleton or array of numeric variables.  These
#  variables define the search space for optimization.  Variables can
#  have initial values, and the value of variable can be retrieved and set.
#
####
#### Var Constructor Syntax
####
#
#  Singleton variable
#
model.x = Var()
#
#  Array of variables
#
model.x = Var(model.A,model.B)
#
#  Option 'initialize' specifies the initial values of variables
#
model.x = Var(initialize=9)
model.x = Var(model.A,initialize={1:1, 2:4, 3:9})
model.x = Var(model.A,initialize=2)
#
#  Option 'initialize' can specify a function used to construct the initial 
#   variable values
#
def f(model, i):
    return 3*i
model.x = Var(model.A,initialize=f)
#
#  Option 'within' specifies a set that is used to constrain variables
#
model.x = Var(within=model.A)
#
#  Option 'bounds' specifies upper and lower bounds for variables.
#  Simple bounds can be specified, or a function that defines bounds for
#  different variables.
#
model.x = Var(bounds=(0.0,1.0))
def f(model, i):
    return (model.x_low[i], model._x_high[i])
model.x = Var(bounds=f)
#
####
#### Var Methods
####
#
# len() - returns the number of variables
#
len(instance.x)
#
# dim() - returns the number of dimensions of the variable index
#
instance.x.dim()
#
# keys() - returns the indices of the variable array
#
instance.x.keys()
#
# reset() - set the variable value with the initial value
#
instance.x.reset()
#
# Coercing variable values explicitly
#
tmp = float(instance.x)
tmp = float(instance.x[i])
#
# Getting variable values (which may be integer or floats)
#
tmp = value(instance.x)
tmp = value(instance.x[i])
tmp = instance.x.value
tmp = instance.x[i].value
#
# Setting variable values
#
instance.x = tmp
instance.x[i] = tmp
#
# NOTE: if x is an array variable, then
#
#  instance.x = tmp
#
# will initialize all of the variable values to the value of 'tmp'.
#
####
#### Var attributes
####
#
# Value - the value of a variable
# Note:  1) value(x) == x.value
#        2) x.value is None until x.reset() is called
#
instance.x.value = 1.0
tmp = instance.x.value
#
# Initial Value - the value of a variable before optimization
#
instance.x.initial = 1.0
#
# Bounds
#
instance.x.setlb(0.0)     # Set a variable lower bound
instance.x.setub(1.0)     # Set a variable upper bound
#
# Fixed - variables that are fixed (and thus not optimized)
#
instance.x.fixed = True     # Fixes this variable value



## -------------------------------------------------------------------------
##
## Objective objects
##
## -------------------------------------------------------------------------
#
#  The Objective object defines an expression that is maximized or minimized
#  during optimization.  The default sense is minimization.  When more than
#  on objective is specified, or when an array of objectives is specified,
#  multi-objective optimizers are needed to simultaneously optimize these
#  objectives.
#
####
#### Objective Constructor Syntax
####
#
#  Singleton objective
#
model.obj = Objective()
#
#  Array of objectives
#
model.obj = Objective(model.A,model.B)
#
#  Option 'rule' can specify a function used to construct the objective
#       expression
#
model.Z = Param(model.A)
model.x = Var(model.A)
def f(model, i):
    return model.Z[i] * model.A[i]
model.obj = Objective(model.A, rule=f)
#
#  Option 'sense' specifies whether the objective is maximized or minimized
#  Note: this option applies to all objectives in an array objective
#
model.obj = Objective(sense=maximize)
#
####
#### Objective Methods
####
#
# len() - returns the number of objectives that are defined
#
len(instance.obj)
#
# dim() - returns the number of dimensions of the objective index
#
instance.obj.dim()
#
# keys() - returns the indices of the objective array
#
instance.obj.keys()
#
####
#### Objective attributes
####
#
# sense - the optimization sense
#
instance.x.sense = maximize
#
# value - returns the value of the objective
# NOTE: this attribute cannot be set
#
tmp = instance.obj.value
tmp = value(instance.obj)
#
# NOTE: if the objective is an array, then this returns a dictionary of
#       objective values, indexed by the array indices.  Thus, the following
#       example computes _all_ objectives, and then returns the objective value
#       for index '2'
#
tmp = instance.obj.value[2]
tmp = value(instance.obj)[2]



## -------------------------------------------------------------------------
##
## Constraint objects
##
## -------------------------------------------------------------------------
#
#  The Constraint object defines an expression that is either bounded or
#  set equal to another expression.  Constraints can be bounded both
#  above and below.
#
####
#### Constraint Constructor Syntax
####
#
#  Singleton constraint
#
model.con = Constraint()
#
#  Array of constraint
#
model.con = Constraint(model.A,model.B)
#
#  Option 'rule' can specify a function used to construct the constraint
#       expression
#
model.Z = Param(model.A)
model.x = Var(model.A)
def f(model, i):
    expr = model.Z[i] * model.A[i]
    return (0, expr, 1)
model.con = Constraint(model.A, rule=f)
#
# Note: the constructor rule must include the specification of bounds
# information for the constraint.  This can be done in one of two ways.  First,
# the rule can return a tuple that includes the values of the lower and 
# upper bounds.  The value 'None' indicates that no bound is specified.
# An equality constraint can be specified as follows:
#
def f(model, i):
    expr = model.Z[i] * model.A[i]
    return (expr, 0)
#
# Second, the constructor rule can augment the expression to include
# bound information.  For example, the previous rule can be rewritten as
#
def f(model, i):
    expr = model.Z[i] * model.A[i]
    expr = expr >= 0
    expr = expr <= 1
    return expr
#
# The following illustrate the type of bounds information that can be
# specified:
#    expr = expr >= val     Lower bound
#    expr = expr > val      Strict lower bound
#    expr = expr < val      Strict upper bound
#    expr = expr <= val     Upper bound
#    expr = expr == val     Equality constraint
#

#
# If the constructor rule returns Constraint.Skip, then the constraint index
# is ignored.  Alternatively, a constructor rule can return a dictionary 
# whose keys define the valid constraint indices.  Thus, the following two
# constraints are equivalent:
#
def f1(model, i):
    if i % 2 == 0:
        return Constraint.Skip
    expr = model.Z[i] * model.A[i]
    return (0, expr, 1)
model.con1 = Constraint(model.A, rule=f1)

def f2(model):
    res = {}
    for i in model.A:
        if i % 2 != 0:
            expr = model.Z[i] * model.A[i]
            res[i] = (0, expr, 1)
    return res
model.con2 = Constraint(model.A, rule=f2)

####
#### Constraint Methods
####
#
# len() - returns the number of constraints that are defined
#
len(instance.con)
#
# dim() - returns the number of dimensions of the constraint index
#
instance.con.dim()
#
# keys() - returns the indices of the constraint array
#
instance.con.keys()
#
####
#### Constraint attributes
####
#
# value - returns the value of the constraint body
# NOTE: this attribute cannot be set
#
tmp = instance.con.value
tmp = value(instance.con)
#
# NOTE: if the constraint is an array, then this returns a dictionary of
#       constraint values, indexed by the array indices.  Thus, the following
#       example computes _all_ constraint, and then returns the constraint value
#       for index '2'
#
tmp = instance.con.value[2]
tmp = value(instance.con)[2]





from coopr.pyomo import *

model = AbstractModel()
#
# Parameter N
#
model.N = Param(within=Integers)
#
# Set I
#
model.I = RangeSet(1,model.N)
#
# Variable b
#
model.b = Var(model.I, domain=Boolean)
#
# Objective zot
#
def costrule(model):
    ans = 0
    for i in model.I:
#               ans += (-1 - .02*i)*model.b[i]
        ans += (1 + .02*i)*model.b[i]
    return ans
#model.zot = Objective(rule=costrule)
model.zot = Objective(rule=costrule, sense=maximize)
#
# Set w_ind
#
def w_ind_rule(model):
    ans = set()
    i = 1
    N9 = model.N.value - 9
    while i <= N9:
        j = i
        i9 = i + 9
        while j <= i9:
            ans.add((i,j))
            j += 1
        i += 1
    return ans
model.w_ind = Set(initialize=w_ind_rule,dimen=2)
#
# Parameter w
#
model.w = Param(model.w_ind)
#
# Set rhs_ind
#
model.rhs_ind = RangeSet(1,model.N-9)
#
# Parameter rhs
#
model.rhs = Param(model.rhs_ind)
#
# Constraint bletch
#
def bletch_rule(model, i):
    ans = 0
    j = i
    i9 = i + 9
    while j <= i9:
        ans += model.w[i,j]*model.b[j]
        j += 1
    ans = ans < model.rhs[i]
    return ans
model.bletch = Constraint(model.rhs_ind, rule=bletch_rule)



#
# Unit Tests for nontrivial Bounds (_SumExpression, _ProductExpression)
#

import os
import sys
from os.path import abspath, dirname
sys.path.insert(0, dirname(dirname(abspath(__file__)))+"/../..")
currdir = dirname(abspath(__file__))+os.sep

import pyutilib.th as unittest
from coopr.pyomo import *
from coopr.opt import *
from nose.tools import nottest

class TestVarSetBounds(unittest.TestCase):

    #Test within=RangeSet()
    def test_rangeset_domain(self):
        if not pyutilib.services.registered_executable("glpsol"):
            return
        self.model = ConcreteModel()
        self.model.s = RangeSet(3) #Set(initialize=[1,2,3])
        self.model.y = Var([1,2], within=self.model.s)
        
        self.model.obj = Objective(expr=self.model.y[1]-self.model.y[2])
        self.model.con1 = Constraint(expr=self.model.y[1] >= 1.1)
        self.model.con2 = Constraint(expr=self.model.y[2] <= 2.9)
        
        self.instance = self.model.create()
        self.opt = SolverFactory("glpk")
        self.results = self.opt.solve(self.instance)
        self.instance.load(self.results)

        self.assertEqual(self.instance.y[1],2)
        self.assertEqual(self.instance.y[2],2)
 

    def test_pyomo_Set_domain(self):
        if not pyutilib.services.registered_executable("glpsol"):
            return
        self.model = ConcreteModel()
        self.model.s = Set(initialize=[1,2,3])
        self.model.y = Var([1,2], within=self.model.s)
        
        self.model.obj = Objective(expr=self.model.y[1]-self.model.y[2])
        self.model.con1 = Constraint(expr=self.model.y[1] >= 1.1)
        self.model.con2 = Constraint(expr=self.model.y[2] <= 2.9)
        
        self.instance = self.model.create()
        self.opt = SolverFactory("glpk")
        self.results = self.opt.solve(self.instance)
        self.instance.load(self.results)

        self.assertEqual(self.instance.y[1],2)
        self.assertEqual(self.instance.y[2],2)
     

    #Bad pyomo Set for variable domain -- empty pyomo Set
    def test_pyomo_Set_domain_empty(self):
        with self.assertRaises(ValueError) as cm:
            self.model = ConcreteModel()
            self.model.s = Set(initialize=[])
            self.model.y = Var([1,2], within=self.model.s)


    #Bad pyomo Set for variable domain -- missing elements
    def test_pyomo_Set_domain_missing(self):
        with self.assertRaises(ValueError) as cm:
            self.model = ConcreteModel()
            self.model.s = Set(initialize=[1,4,5])
            self.model.y = Var([1,2], within=self.model.s)


    #Bad pyomo Set for variable domain -- noninteger elements
    def test_pyomo_Set_domain_duplicates(self):
        with self.assertRaises(ValueError) as cm:
            self.model = ConcreteModel()
            self.model.s = Set(initialize=[1.7,2,3])
            self.model.y = Var([1,2], within=self.model.s)
 

    def test_pyomo_Set_dat_file_domain(self):
        if not pyutilib.services.registered_executable("glpsol"):
            return
        self.model = AbstractModel()
        self.model.s = Set()
        self.model.y = Var([1,2], within=self.model.s)
       
        def obj_rule(model):
            return sum(model.y[i]*(-1)**(i-1) for i in model.y)
        self.model.obj = Objective(rule=obj_rule) #sum(self.model.y[i]*(-1)**(i-1) for i in self.model.y))
        self.model.con = Constraint([1,2],expr=lambda model, i : model.y[i]*(-1)**(i-1) >= (1.1)**(2-i) * (-2.9)**(i-1))
        
        self.instance = self.model.create(currdir+"vars_dat_file.dat")
        self.opt = SolverFactory("glpk")
        self.results = self.opt.solve(self.instance)
        self.instance.load(self.results)

        self.assertEqual(self.instance.y[1],2)
        self.assertEqual(self.instance.y[2],2)
     

    #Bad pyomo Set for variable domain -- empty pyomo Set
    def test_pyomo_Set_dat_file_domain_empty(self):
        with self.assertRaises(ValueError) as cm:
            self.model = AbstractModel()
            self.model.s = Set()
            self.model.y = Var([1,2], within=self.model.s)
            self.instance = self.model.create(currdir+"vars_dat_file_empty.dat")


    #Bad pyomo Set for variable domain -- missing elements
    def test_pyomo_Set_dat_file_domain_missing(self):
        with self.assertRaises(ValueError) as cm:
            self.model = AbstractModel()
            self.model.s = Set()
            self.model.y = Var([1,2], within=self.model.s)
            self.instance = self.model.create(currdir+"vars_dat_file_missing.dat")


    #Bad pyomo Set for variable domain -- noninteger elements
    def test_pyomo_Set_dat_file_domain_duplicates(self):
        with self.assertRaises(ValueError) as cm:
            self.model = AbstractModel()
            self.model.s = Set()
            self.model.y = Var([1,2], within=self.model.s)
            self.instance = self.model.create(currdir+"vars_dat_file_nonint.dat")
 

    #Test within=list -- this works for range() since range() returns a list
    def test_list_domain(self):
        if not pyutilib.services.registered_executable("glpsol"):
            return
        self.model = ConcreteModel()
        self.model.y = Var([1,2], within=[1,2,3])
        
        self.model.obj = Objective(expr=self.model.y[1]-self.model.y[2])
        self.model.con1 = Constraint(expr=self.model.y[1] >= 1.1)
        self.model.con2 = Constraint(expr=self.model.y[2] <= 2.9)
        
        self.instance = self.model.create()
        self.opt = SolverFactory("glpk")
        self.results = self.opt.solve(self.instance)
        self.instance.load(self.results)

        self.assertEqual(self.instance.y[1],2)
        self.assertEqual(self.instance.y[2],2)


    #Bad list for variable domain -- empty list
    def test_list_domain_empty(self):
        with self.assertRaises(ValueError) as cm:
            self.model = ConcreteModel()
            self.model.y = Var([1,2], within=[])


    #Bad list for variable domain -- missing elements
    def test_list_domain_bad_missing(self):
        with self.assertRaises(ValueError) as cm:
            self.model = ConcreteModel()
            self.model.y = Var([1,2], within=[1,4,5])


    #Bad list for variable domain -- duplicate elements
    def test_list_domain_bad_duplicates(self):
        with self.assertRaises(ValueError) as cm:
            self.model = ConcreteModel()
            self.model.y = Var([1,2], within=[1,1,2,3])


    #Bad list for variable domain -- noninteger elements
    def test_list_domain_bad_duplicates(self):
        with self.assertRaises(ValueError) as cm:
            self.model = ConcreteModel()
            self.model.y = Var([1,2], within=[1.7,2,3])
   

    #Test within=set() -- python native set, not pyomo Set object
    def test_set_domain(self):
        if not pyutilib.services.registered_executable("glpsol"):
            return
        self.model = ConcreteModel()
        self.model.y = Var([1,2], within=set([1,2,3]))
        
        self.model.obj = Objective(expr=self.model.y[1]-self.model.y[2])
        self.model.con1 = Constraint(expr=self.model.y[1] >= 1.1)
        self.model.con2 = Constraint(expr=self.model.y[2] <= 2.9)
        
        self.instance = self.model.create()
        self.opt = SolverFactory("glpk")
        self.results = self.opt.solve(self.instance)
        self.instance.load(self.results)

        self.assertEqual(self.instance.y[1],2)
        self.assertEqual(self.instance.y[2],2)


    #Bad set for variable domain -- empty set
    def test_set_domain_empty(self):
        with self.assertRaises(ValueError) as cm:
            self.model = ConcreteModel()
            self.model.y = Var([2,2], within=set([]))


    #Bad set for variable domain -- missing elements
    def test_set_domain_bad_missing(self):
        with self.assertRaises(ValueError) as cm:
            self.model = ConcreteModel()
            self.model.y = Var([1,2], within=set([1,4,5]))


    #Bad set for variable domain -- duplicate elements
    def test_set_domain_bad_duplicates(self):
        with self.assertRaises(ValueError) as cm:
            self.model = ConcreteModel()
            self.model.y = Var([1,2], within=set([1,1,2,3]))


    #Bad set for variable domain -- noninteger elements
    def test_set_domain_bad_duplicates(self):
        with self.assertRaises(ValueError) as cm:
            self.model = ConcreteModel()
            self.model.y = Var([1,2], within=set([1.7,2,3]))
   

    #Test within=xrange()
    def test_rangeset_domain(self):
        if not pyutilib.services.registered_executable("glpsol"):
            return
        self.model = ConcreteModel()
        self.model.y = Var([1,2], within=xrange(4))
        
        self.model.obj = Objective(expr=self.model.y[1]-self.model.y[2])
        self.model.con1 = Constraint(expr=self.model.y[1] >= 1.1)
        self.model.con2 = Constraint(expr=self.model.y[2] <= 2.9)
        
        self.instance = self.model.create()
        self.opt = SolverFactory("glpk")
        self.results = self.opt.solve(self.instance)
        self.instance.load(self.results)

        self.assertEqual(self.instance.y[1],2)
        self.assertEqual(self.instance.y[2],2)





"""
       / -9x-20, -5 <= x <= -4
       | -7x-12, -4 <= x <= -3
       | -5x-6 , -3 <= x <= -2
       | -3x-2 , -2 <= x <= -1
f(x) = | 1     , -1 <= x <=  1 
       | 3x-2  ,  1 <= x <=  2
       | 5x-6  ,  2 <= x <=  3
       \ 7x-12 ,  3 <= x <=  4

"""

from coopr.pyomo import *

INDEX_SET = range(1,8) # There will be two copies of this function
DOMAIN_PTS = dict([(t,[float(i) for i in (range(-5,0)+range(1,5))]) for t in INDEX_SET])

 
def F(model,t,x):
    return (x**2)*model.p[t]

def define_model(**kwds):

    model = ConcreteModel()

    model.x = Var(INDEX_SET, bounds=(-5,4)) # domain variable
    model.Fx = Var(INDEX_SET) # range variable
    model.p = Param(INDEX_SET,initialize=1.0)

    model.obj = Objective(expr=summation(model.Fx), sense=kwds.pop('sense',maximize))

    model.piecewise = Piecewise(INDEX_SET,model.Fx,model.x,
                                  pw_pts=DOMAIN_PTS,
                                  f_rule=F, **kwds)

    #Fix the answer for testing purposes
    model.set_answer_constraint1 = Constraint(expr= model.x[1] == -5.0)
    model.set_answer_constraint2 = Constraint(expr= model.x[2] == -3.0)
    model.set_answer_constraint3 = Constraint(expr= model.x[3] == -2.5)
    model.set_answer_constraint4 = Constraint(expr= model.x[4] == -1.5)
    model.set_answer_constraint5 = Constraint(expr= model.x[5] == 2.0)
    model.set_answer_constraint6 = Constraint(expr= model.x[6] == 3.5)
    model.set_answer_constraint7 = Constraint(expr= model.x[7] == 4.0)    

    return model



#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the Coopr README.txt file.
#  _________________________________________________________________________

import os.path
import re
from TableData import TableData
from pyutilib.component.core import alias


class TextTable(TableData):

    alias("tab", "Manage IO with ASCI tables.")

    def __init__(self):
        TableData.__init__(self)

    def open(self):
        if self.filename is None:
            raise IOError, "No filename specified"
        if not os.path.exists(self.filename):
            raise IOError, "Cannot find file '%s'" % self.filename
        self.INPUT = open(self.filename, 'r')

    def close(self):
        self.INPUT.close()

    def read(self):
        tmp=[]
        for line in self.INPUT:
            line=line.strip()
            tokens = re.split("[\t ]+",line)
            if tokens != ['']:
                tmp.append(tokens)
        if len(tmp) == 0:
            raise IOError, "Empty *.tab file"
        elif len(tmp) == 1:
            if not self.options.param is None:
                self._info = ["param",self.options.param,":=",tmp[0][0]]
            elif len(self.options.symbol_map) == 1:
                self._info = ["param",self.options.symbol_map[self.options.symbol_map.keys()[0]],":=",tmp[0][0]]
            else:
                raise IOError, "Data looks like a parameter, but multiple parameter names have been specified: %s" % str(self.options.symbol_map)
        else:
            self._set_data(tmp[0], tmp[1:])
        return True



import pickle
from coopr.pyomo import *
from pyutilib.misc import Options
options = Options()
options.solver = "cplex"
options.solver_io = "lp"
options.keepfiles = True
#options.tee = True
options.debug = True
options.symbolic_solver_labels=True

kwds = {'pw_constr_type':'EQ','pw_repn':'SOS2','sense':maximize,'force_pw':True}

#problem_name = "piecewise_multi_vararray"
#problem_name = "concave_multi_vararray1"
#problem_name = "concave_multi_vararray2"
#problem_name = "convex_multi_vararray1"
#problem_name = "convex_multi_vararray2"
#problem_name = "convex_vararray"
#problem_name = "concave_vararray"
#problem_name = "convex_var"
problem_name = "concave_var"
#problem_name = "piecewise_var"
#problem_name = "piecewise_vararray"

p = __import__(problem_name)

model = p.define_model(**kwds)
inst = model.create()

print
print
for block in inst.all_blocks():
    print block.components.components(ctype=Block)
inst.pprint()
print
print

data = Options(options=options)
opt_data = scripting.util.apply_optimizer(data,instance=inst)

results = opt_data.results

inst.load(results)

res = dict([(v[idx].name,v[idx].value) for (n,v) in inst.active_components(Var).iteritems() for idx in v if (n[:2] == 'Fx') or (n[:1] == 'x')])
print res

f = open(problem_name+'_baseline_results.pickle','w')
pickle.dump(res,f)
f.close()



#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the Coopr README.txt file.
#  _________________________________________________________________________

# TODO
# . rename 'filter' to something else
# . confirm that filtering is efficient

__all__ = ['Set', '_SetContainer', '_SetArray', 'set_options', 'simple_set_rule', 'SetOf']

import logging
import sys
import types
import copy
import itertools
try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict
try:
    import bidict
    using_bidict=True
except:
    using_bidict=False

import pyutilib.misc
from pyutilib.component.core import alias
from misc import apply_indexed_rule, apply_parameterized_indexed_rule
from component import Component


logger = logging.getLogger('coopr.pyomo')


def set_options(**kwds):
    """
    This is a decorator for set initializer functions.  This
    decorator allows an arbitrary dictionary of values to passed
    through to the set constructor.

    Example:
    TODO
    """
    def decorator(func):
        func.set_options = kwds
        return func
    return decorator


def simple_set_rule( fn ):
    """This is a decorator that translates None into Set.End.
    This supports a simpler syntax in set rules, though these can be
    more difficult to debug when errors occur.

    Example:

    @simple_set_rule
    def A_rule(model, i, j):
        ...
    """

    def wrapper_function ( *args, **kwargs ):
        value = fn( *args, **kwargs )
        if value is None:
            return Set.End
        return value
    return wrapper_function


class Set(Component):
    """Set objects that are used to index other Pyomo objects

       This class has a similar look-and-feel as a built-in set class.  However,
       the set operations defined in this class return another abstract
       Set object.  This class contains a concrete set, which can be
       initialized by the load() method.
    
    Constructor
        Arguments:
           name          The name of the set
           within        A set that defines the type of values that can
                          be contained in this set
           rule
           initialize    Default set members, which may be overriden
                          when setting up this set
           validate      Define a function for validating membership in a
                          set.  This has the functional form:
                          f: data -> bool
                          and returns true if the data belongs in the set
           dimen         Specify the set's arity.
           doc           Documentation for this object
           virtual       Indicate that this is a virtual set that does not
                            have concrete members.
    """

    alias('Set', doc="Set data that is used to define a model instance.", subclass=True)

    End             = (1003,)

    def __new__(cls, *args, **kwds):
        if cls != Set:
            return super(Set, cls).__new__(cls)
        if args == ():
            return _SetContainer.__new__(_SetContainer)
        else:
            return _SetArray.__new__(_SetArray)

    def __init__(self, **kwds):
        self.order=[]
        # IMPT: order_dict keys are 1-based.
        if using_bidict:
            self.order_dict = bidict.bidict()
        else:
            self.order_dict = {}
            self.order_dict_inv = {}

        # 'rule' and 'initialize' are mutually exclusive options
        if "rule" in kwds and "initialize" in kwds:
            raise TypeError, "Cannot specify both 'rule' and 'initialize' " + \
                  "keywords in '%s'" % str(self)

        # Get keyword arguments
        kwds.setdefault("name", "_unknown_")
        self.initialize = kwds.pop("rule", None)
        self.initialize = kwds.pop("initialize", self.initialize)
        self.validate   = kwds.pop("validate", None)
        self.ordered    = kwds.pop("ordered", False)
        self.virtual    = kwds.pop("virtual", False)
        self.concrete   = not self.virtual
        self._bounds    = kwds.pop("bounds", None)
        self.filter     = kwds.pop("filter", None)

        # We can't access self.dimen after its been written, so we use
        # tmp_dimen until the end of __init__
        tmp_dimen = None

        # Get dimen from within, if possible
        self.domain = kwds.pop("within", None)
        if self.domain is not None:
            tmp_dimen = self.domain.dimen
        if self._bounds is None and not self.domain is None:
            self._bounds = copy.copy(self.domain._bounds)

        # Make sure dimen and implied dimensions don't conflict
        kwd_dimen = kwds.pop("dimen", None)
        if kwd_dimen is not None:
            if self.domain is not None and tmp_dimen != kwd_dimen:
                raise ValueError, \
                      ("Value of keyword 'dimen', %s, differs from the " + \
                       "dimension of the superset '%s', %s") % \
                       (str(kwd_dimen), str(self.domain.name), str(tmp_dimen))
            else:
                tmp_dimen = kwd_dimen

        kwds.setdefault('ctype', Set)
        Component.__init__(self, **kwds)

        if tmp_dimen is None:
            # We set the default to 1
            tmp_dimen = 1
        if self.initialize is not None:
            #
            # Convert generators to lists, since a generator is not copyable
            #
            if type(self.initialize) is types.GeneratorType:
                self.initialize = list(self.initialize)
            #
            # Try to guess dimen from the initialize list
            #
            tmp=None
            if type(self.initialize) is tuple:
                tmp = len(self.initialize)
            elif type(self.initialize) is list and len(self.initialize) > 0 \
                     and type(self.initialize[0]) is tuple:
                tmp = len(self.initialize[0])
            else:
                tmp = getattr(self.initialize, 'dimen', tmp)
            if tmp is not None:
                if kwd_dimen is not None and tmp != kwd_dimen:
                    raise ValueError, "Dimension argument differs from the data in the initialize list"
                else:
                    tmp_dimen = tmp

        self.dimen = tmp_dimen

    def valid_model_component(self):
        if self.virtual and not self.concrete:
            return False
        return True

    def reset(self):
        pass

    def dim(self):
        return self._ndim

    def display(self, ostream=None, verbose=False):
        self.pprint(ostream=ostream, verbose=verbose)

    def _verify(self,element,use_exception=True):
        # A utility routine that is used to verify if the element
        # is valid for this set.
        #print "HERE VERIFYING",element,self.validate,self.domain,self._model
        if self.domain is not None and element not in self.domain:
            if use_exception:
                raise ValueError, "The value="+str(element)+" is not valid for set="+self.name+", because it is not within the domain="+self.domain.name
            return False
        if self.validate is not None:
            if self._parent is not None:
                flag = apply_indexed_rule(self, self.validate, self._parent(), element)
            else:
                flag = apply_indexed_rule(self, self.validate, None, element)
            if not flag:
                if use_exception:
                    raise ValueError, "The value="+str(element)+" violates the validation rule of set="+self.name
                return False
        if self.dimen > 1 and type(element) is not tuple:
            if use_exception:
                raise ValueError, "The value="+str(element)+" is not a tuple for set="+self.name+", which has dimen="+str(self.dimen)
            return False
        elif self.dimen == 1 and type(element) is tuple:
            if use_exception:
                raise ValueError, "The value="+str(element)+" is a tuple for set="+self.name+", which has dimen="+str(self.dimen)
            return False
        elif type(element) is tuple and len(element) != self.dimen:
            if use_exception:
                raise ValueError, "The value="+str(element)+" does not have dimension="+str(self.dimen)+", which is needed for set="+self.name
            return False
        return True

    def bounds(self):
        """Return bounds information.  The default value is 'None', which
        indicates that this set does not contain bounds."""
        return self._bounds



class _SetContainer(Set):
    """A derived Set object that contains a single set."""

    def __init__(self, *args, **kwds):
        """ Constructor """
        if args != ():
            raise TypeError, "A _SetContainer expects no arguments"
        self._index_set=None
        self._index=[None]
        self._ndim=0
        self.value=set()
        Set.__init__(self, **kwds)

    def construct(self, values=None):
        """ Apply the rule to construct values in this set """
        if __debug__:
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug("Constructing _SetContainer, name="+self.name+", from data="+repr(values))
        if self._constructed:
            return
        self._constructed=True

        if self.initialize is None:
            self.initialize = getattr(self,'rule',None)
        #
        # Construct using the values list
        #
        if values is not None:
            if type(self._bounds) is tuple:
                first=self._bounds[0]
                last=self._bounds[1]
            else:
                first=None
                last=None
            all_numeric=True
            for val in values[None]:
                if type(val) in [int,float,long]:
                    if first is None or val<first:
                        first=val
                    if last is None or val>last:
                        last=val
                else:
                    all_numeric=False
                self.add(val)
            if all_numeric:
                self._bounds = (first,last)
        #
        # Construct using the rule
        #
        elif type(self.initialize) is types.FunctionType:
            if self._parent is None:
                raise ValueError, "Must pass the parent block in to initialize with a function"
            if self.initialize.func_code.co_argcount == 1:
                #
                # Using a rule of the form f(model) -> iterator
                #
                tmp = self.initialize(self._parent())
                for val in tmp:
                    if self.dimen is None:
                        if type(val) in [tuple,list]:
                            self.dimen=len(val)
                        else:
                            self.dimen=1
                    self.add(val)
            else:
                #
                # Using a rule of the form f(model, z) -> element
                #
                ctr=1
                val = apply_indexed_rule(self, self.initialize, self._parent(), ctr)
                if val is None:
                    #logger.warning("DEPRECATION WARNING: Set rule returned None instead of Set.Skip")
                    raise ValueError, "Set rule returned None instead of Set.Skip"
                if self.dimen is None:
                    if type(val) in [tuple,list] and not val == Set.End:
                        self.dimen=len(val)
                    else:
                        self.dimen=1
                while not (val.__class__ is tuple and val == Set.End):
                    self.add(val)
                    ctr += 1
                    val = apply_indexed_rule(self, self.initialize, self._parent(), ctr)
                    if val is None:
                        #logger.warning("DEPRECATION WARNING: Set rule returned None instead of Set.Skip")
                        raise ValueError, "Set rule returned None instead of Set.Skip"
        #
        # Construct using the default values
        #
        elif self.initialize is not None:
            if type(self.initialize) is dict:
                raise ValueError, "Cannot initialize set "+self.name+" with dictionary data"
            if type(self._bounds) is tuple:
                first=self._bounds[0]
                last=self._bounds[1]
            else:
                first=None
                last=None
            all_numeric=True
            for val in self.initialize:
                if type(val) in [int,float,long]:
                    if first is None or val<first:
                        first=val
                    if last is None or val>last:
                        last=val
                else:
                    all_numeric=False
                self.add(val)
            #print "HERE",self.name,first,last,all_numeric
            if all_numeric:
                self._bounds = (first,last)

        #
        # Apply filter, if defined
        #
        if self.filter is not None:
            self._apply_filter()

    def _apply_filter(self):
        """
        Applies self.filter to each element. If self.filter returns True,
        the element remain in the Set; otherwise, it is removed.
        """
        toRemove = []
        for el in self.value:
            flag = apply_indexed_rule(self, self.filter, self._parent(), el)
            if not flag:
                toRemove.append(el)
        for el in toRemove:
            self.remove(el)

    def data(self):
        """The underlying set data."""
        if not self.concrete:
            raise TypeError, "Cannot access underlying set data for a non-concrete set "+self.name
        return self.value

    def clear(self):
        """Remove all elements from the set."""
        if self.virtual:
            raise TypeError, "Cannot clear virtual Set object `"+self.name+"'"
        self.value.clear()
        if using_bidict:
            self.order_dict = bidict.bidict()
        else:
            self.order_dict = {}
            self.order_dict_inv = {}

    def check_values(self):
        """ Verify that the values in this set are valid.
        """
        if self.virtual:
            return
        for val in self.value:
            self._verify(val)

    def add(self, *args):
        """Add one or more elements to a set."""
        if self.virtual:
            raise TypeError, "Cannot add elements to virtual set `"+self.name+"'"
        for val in args:
            tmp = pyutilib.misc.flatten_tuple(val)
            self._verify(tmp)
            try:
                if tmp in self.value:
                    continue
                    #raise ValueError, "Element "+str(tmp)+" already exists in set "+self.name
                self.value.add(tmp)
                if self.ordered:
                    if tmp in self.order_dict:
                        raise ValueError, "Element "+str(tmp)+" already exists in ordered set "+self.name
                    self.order.append(tmp)
                    if using_bidict:
                        self.order_dict[tmp:] = len(self.order)
                    else:
                        self.order_dict[tmp] = len(self.order)
                        self.order_dict_inv[len(self.order)] = tmp


            except TypeError:
                raise TypeError, "Problem inserting "+str(tmp)+" into set "+self.name

    def remove(self, element):
        """Remove an element from the set.
           If the element is not a member, raise an error.
        """
        if self.virtual:
            raise KeyError, "Cannot remove element `"+str(element)+"' from virtual set "+str(self.name)
        if element not in self.value:
            raise KeyError, "Cannot remove element `"+str(element)+"' from set "+str(self.name)
        self.value.remove(element)
        if self.ordered:
            id = self.order_dict[element]-1
            for i in xrange(id+1,len(self.order)):
                if using_bidict:
                    self.order_dict[self.order[i]] = i
                else:
                    self.order_dict[self.order[i]] = i
                    self.order_dict_inv[i] = self.order[i]
            del self.order[id]

    def discard(self, element):
        """Remove an element from the set.
           If the element is not a member, do nothing.
        """
        if self.virtual:
            raise KeyError, "Cannot discard element `"+str(element)+"' from virtual set "+str(self.name)
        if self.ordered and element in self.value:
            self.order.remove(element)
        self.value.discard(element)

    def first(self):
        if not self.concrete:
            raise TypeError, "Cannot access the first element of a non-concrete set `"+self.name+"'"
        return self.member(1)

    def last(self):
        if self.virtual:
            raise TypeError, "Cannot access the last element of a non-concrete set `"+self.name+"'"
        return self.member(len(self.value))

    def __getitem__(self, key):
        return self.member(key)

    def member(self, key):
        if not self.concrete:
            raise TypeError, "Cannot access a specific element of a non-concrete set `"+self.name+"'"
        if self.ordered:
            if using_bidict:
                return self.order_dict.inv[key]
            else:
                return self.order_dict_inv[key]
        #
        # If the set is not ordered, then we use the intrinsic ordering imposed by the set.
        # We convert the set to a list and return the specified value.
        #
        if key >= 1:
            return list(self.value)[key-1]
        elif key < 0:
            return list(self.value)[key]
        else:
            raise IndexError, "Valid index values for sets are 1 ... len(set) or -1 ... -len(set)"

    def ord(self, match_element):
        """ For ordered sets, return the position index (1-based)
            of the input element.
        """
        if self.ordered is False:
            raise AttributeError, "Cannot invoke ord() method for unordered set="+str(self.name)
        try:
            return self.order_dict[match_element]
        except IndexError:
            raise IndexError, "Unknown input element="+str(match_element)+" provided as input to ord() method for set="+str(self.name)

    def next(self, match_element, k=1):
        # NOTE: we currently aren't consist (and don't actually deal with) w.r.t. unordered sets - see the implementation of member() above.
        try:
            element_position = self.order_dict[match_element]
        except KeyError:
            raise KeyError, "Cannot obtain next() member of set="+self.name+"; input element="+str(match_element)+" is not a member of the set!"

        try:
            return self.order[element_position+k-1] # "order" is a list, accessed 0-based.
        except KeyError:
            raise KeyError, "Cannot obtain next() member of set="+self.name+"; failed to access item in position="+str(element_position+k-1)

    def nextw(self, match_element, k=1):
        ndx = self.order_dict[match_element]+k
        total = len(self.order)
        if ndx > total:
            ndx -= total
        if ndx <= 0:
            ndx += total
        return self.order.inv[ndx]

    def prev(self, match_element, k=1):
        return self.next(match_element, k=-k)

    def prevw(self, match_element, k=1):
        return self.nextw(match_element, k=-k)

    def pprint(self, ostream=None, verbose=False):
        if ostream is None:
            ostream = sys.stdout
        print >>ostream, "  ",self.name,":",
        if not self.doc is None:
            print >>ostream, self.doc
            print >>ostream, "  ",
        print >>ostream, "\tDim="+str(self.dim()),
        print >>ostream, "\tDimen="+str(self.dimen),
        print >>ostream, "\tSize="+str(len(self)),
        if self.domain is not None:
            print >>ostream, "\tDomain="+str(self.domain.name),
        else:
            print >>ostream, "\tDomain="+str(None),
        print >>ostream, "\tOrdered="+str(self.ordered),
        print >>ostream, "\tBounds="+str(self._bounds)
        if self._model is None:
            print >>ostream, "\t Model=None"
        else:
            print >>ostream, "\t Model="+str(self._model().name)
        if not self.concrete or (self.virtual and not verbose):
            print >>ostream, "\t  Virtual"
        else:
            if self.ordered:
                tmp = copy.copy(self.order)
            else:
                tmp = copy.copy(list(self.data()))
                tmp.sort()
            print >>ostream, "\t  ",tmp

    def __len__(self):
        """ The number of items in the set.
            The set is empty until a concrete instance has been setup.
        """
        if not self.concrete:
            raise ValueError, "The size of a non-concrete set is unknown"
        return len(self.data())

    def __iter__(self):
        """Return an iterator for the underlying set"""
        if self.virtual:
            raise TypeError, "Cannot iterate over Set object `"+self.name+"', which is a virtual set"
        if self.ordered:
            return self.order.__iter__()
        return self.value.__iter__()

    def __reversed__(self):
        return reversed(self.__iter__())

    def __hash__(self):
        """Hash this object"""
        return Set.__hash__(self)

    def __eq__(self,other):
        """ Equality comparison """
        if id(self) == id(other):
            return True
        if other is None:
            return False
        try:
            tmp = self._set_repn(other)
        except:
            return False
        if self.virtual:
            if tmp.virtual:
                return hash(self) == hash(tmp)
            return False
        if tmp.virtual:
            return False
        #print 'x', self.virtual, other.virtual, tmp.virtual
        if self.dimen != other.dimen:
            return False
        return self.value.__eq__( tmp.data() )

    def __ne__(self,other):
        return not self.__eq__(other)

    def __contains__(self, element):
        """Report whether an element is a member of this set.
        (Called in response to the expression 'element in self'.)
        """
        #
        # If the element is a set, then see if this is a subset
        #
        # NB: since 95+% of all sets are indexed by int's (or tuples of
        # ints) this is a quick test that can avoid (expensive)
        # isinstance calls that usually return "False".
        #
        element_t = type(element)
        if element_t is not int and element_t is not tuple and \
               isinstance(element,_SetContainer):
            return element.issubset(self)
        # 
        # When dealing with a concrete set, just check if the element is
        # in the set. there is no need for extra validation.
        #        
        if self._constructed and self.concrete is True:
           return self._set_contains(element)
        #
        # If this is not a valid element, then return False
        #
        if not self._verify(element,False):
            return False
        #
        # If the restriction rule is used then we do not actually
        # check whether the data is in the set self.value.
        #
        if self.validate is not None and not self.concrete:
            return True
        #
        # The final check: return true if self.concrete is False, since we should
        # have already validated this value.  The following, or at least one of
        # the execution paths - is probably redundant with the above.
        #
        return not self.concrete or self._set_contains(element)

    def _set_contains(self, element):
        return element in self.value

    def isdisjoint(self, other):
        """Return True if the set has no elements in common with other. Sets are disjoint if and only if their intersection is the empty set."""
        other = self._set_repn(other)
        tmp = self & other
        for elt in tmp:
            return False
        return True
        
    def issubset(self,other):
        """Report whether another set contains this set"""
        other = self._set_repn(other)
        if self.dimen != other.dimen:
            raise ValueError, "Cannot perform set operation with sets "+self.name+" and "+other.name+" that have different element dimensions: "+str(self.dimen)+" "+str(other.dimen)
        if not self.concrete:
            raise TypeError, "ERROR: cannot perform \"issubset\" test because the current set is not a concrete set."
        for val in self:
            if val not in other:
                return False
        return True

    def __lt__(self,other):
        return self <= other and not self == other

    def issuperset(self, other):
        """Report this set contains another set"""
        other = self._set_repn(other)
        if self.dimen != other.dimen:
            raise ValueError, "Cannot perform set operation with sets "+self.name+" and "+other.name+" that have different element dimensions: "+str(self.dimen)+" "+str(other.dimen)
        if not other.concrete:
            raise TypeError, "ERROR: cannot perform \"issuperset\" test because the target set is not a concrete set."
        for val in other:
            if val not in self:
                return False
        return True

    def __gt__(self,other):
        return self >= other and not self == other

    __le__ = issubset
    __ge__ = issuperset

    def _set_repn(self, other):
        if isinstance(other,_SetContainer):
            return other
        return SetOf(other)

    def union(self, *args):
        """Return the union of this set with one or more sets"""
        tmp = self
        for arg in args:
            tmp = _SetUnion(tmp, arg)
        return tmp

    def __or__(self,other):
        """Return the union of this set with another set"""
        return self.union(other)

    def intersection(self, *args):
        """Return the intersection of this set with one or more sets"""
        tmp = self
        for arg in args:
            tmp = _SetIntersection(tmp, arg)
        return tmp

    def __and__(self,other):
        """Return the intersection of this set with another set"""
        return self.intersection(other)

    def difference(self, *args):
        """Return the difference between this set with one or more sets"""
        tmp = self
        for arg in args:
            tmp = _SetDifference(tmp, arg)
        return tmp

    def __sub__(self,other):
        """Return the difference between this set and another set"""
        return self.difference(other)

    def symmetric_difference(self, *args):
        """Return the symmetric difference of this set with one or more sets"""
        tmp = self
        for arg in args:
            tmp = _SetSymmetricDifference(tmp, arg)
        return tmp

    def __xor__(self,other):
        """Return the symmetric difference of this set with another set"""
        return self.symmetric_difference(other)

    def cross(self, *args):
        """Return a cross-product between this set and one or more sets"""
        tmp = self
        for arg in args:
            tmp = _SetProduct(tmp, arg)
        return tmp

    def __mul__(self,other):
        """Return a cross-product between this set and another set"""
        return self.cross(other)


class SetOf(_SetContainer):
    """A derived _SetContainer object that creates a set from external data without duplicating it."""

    alias('SetOf', "Define a Pyomo Set component using an iterable data object.")

    def __init__(self, *args, **kwds):
        if len(args) != 1:
            raise TypeError, "Only one set data argument can be specified"
        _SetContainer.__init__(self,**kwds)
        self.value = None
        self._constructed = True
        self.virtual = True
        self.concrete = True
        #
        self._data = args[0]
        try:
            for i in self._data:
                if type(i) is tuple:
                    self.dimen = len(i)
                else:
                    self.dimen = 1
                break
        except TypeError, e:
            raise TypeError, "Cannot create a Pyomo set: "+str(e)

    def construct(self, values=None):
        """Disabled construction method"""
        pass

    def __len__(self):
        """The number of items in the set."""
        try:
            return len(self._data)
        except:
            pass
        ctr = 0
        for i in self:
            ctr += 1
        return ctr

    def __iter__(self):
        """Return an iterator for the underlying set"""
        for i in self._data:
            yield i

    def _set_contains(self, element):
        return element in self._data
        
    def data(self):
        """The underlying set data."""
        return set(self)

    def __eq__(self, other):
        """ Equality comparison """
        if other is None:
            return False
        tmp = self._set_repn(other)
        if self.dimen != other.dimen:
            return False
        ctr = 0
        for i in self:
            if not i in other:
                return False
            ctr += 1
        return ctr == len(other)
            

class _SetOperator(_SetContainer):
    """A derived _SetContainer object that contains a concrete virtual single set."""

    def __init__(self, *args, **kwds):
        if len(args) != 2:
            raise TypeError, "Two arguments required for a binary set operator"
        dimen_test = kwds.get('dimen_test',True)
        if 'dimen_test' in kwds:
            del kwds['dimen_test']
        _SetContainer.__init__(self,**kwds)
        self.value = None
        self._constructed = True
        self.virtual = True
        self.concrete = True
        #
        self._setA = args[0]
        if not self._setA.concrete:
            raise TypeError, "Cannot perform set operations with non-concrete set '"+self._setA.name+"'"
        if isinstance(args[1],_SetContainer):
            self._setB = args[1]
        else:
            self._setB = SetOf(args[1])
        if not self._setB.concrete:
            raise TypeError, "Cannot perform set operations with non-concrete set '"+self._setB.name+"'"
        if dimen_test and self._setA.dimen != self._setB.dimen:
            raise ValueError, "Cannot perform set operation with sets "+self._setA.name+" and "+self._setB.name+" that have different element dimensions: "+str(self._setA.dimen)+" "+str(self._setB.dimen)
        self.dimen = self._setA.dimen
        #
        self.ordered = self._setA.ordered and self._setB.ordered

    def construct(self, values=None):
        """ Disabled construction method """
        pass

    def __len__(self):
        """The number of items in the set."""
        ctr = 0
        for i in self:
            ctr += 1
        return ctr

    def __iter__(self):
        """Return an iterator for the underlying set"""
        raise IOError, "Undefined set iterator"

    def _set_contains(self, element):
        raise IOError, "Undefined set operation"
        
    def data(self):
        """The underlying set data."""
        return set(self)

    def __eq__(self, other):
        """ Equality comparison """
        if other is None:
            return False
        tmp = self._set_repn(other)
        if self.dimen != other.dimen:
            return False
        ctr = 0
        for i in self:
            if not i in other:
                return False
            ctr += 1
        return ctr == len(other)
            

class _SetUnion(_SetOperator):

    def __init__(self, *args, **kwds):
        _SetOperator.__init__(self, *args, **kwds)

    def __iter__(self):
        for elt in self._setA:
            yield elt
        for elt in self._setB:
            if not elt in self._setA:
                yield elt

    def _set_contains(self, elt):
        return elt in self._setA or elt in self._setB
    

class _SetIntersection(_SetOperator):

    def __init__(self, *args, **kwds):
        _SetOperator.__init__(self, *args, **kwds)

    def __iter__(self):
        for elt in self._setA:
            if elt in self._setB:
                yield elt

    def _set_contains(self, elt):
        return elt in self._setA and elt in self._setB
    

class _SetDifference(_SetOperator):

    def __init__(self, *args, **kwds):
        _SetOperator.__init__(self, *args, **kwds)

    def __iter__(self):
        for elt in self._setA:
            if not elt in self._setB:
                yield elt

    def _set_contains(self, elt):
        return elt in self._setA and not elt in self._setB
    

class _SetSymmetricDifference(_SetOperator):

    def __init__(self, *args, **kwds):
        _SetOperator.__init__(self, *args, **kwds)

    def __iter__(self):
        for elt in self._setA:
            if not elt in self._setB:
                yield elt
        for elt in self._setB:
            if not elt in self._setA:
                yield elt

    def _set_contains(self, elt):
        return elt in self._setA ^ elt in self._setB
    

class _SetProduct(_SetOperator):

    def __init__(self, *args, **kwd):
        kwd['dimen_test'] = False
        _SetOperator.__init__(self, *args, **kwd)
        # the individual index sets definining the product set.
        if isinstance(self._setA,_SetProduct):
            self.set_tuple = self._setA.set_tuple
        else:
            self.set_tuple = [self._setA]
        if isinstance(self._setB,_SetProduct):
            self.set_tuple += self._setB.set_tuple
        else:
            self.set_tuple.append(self._setB)
        # set the "dimen" instance attribute.
        self._compute_dimen()

    def __iter__(self):
        if self.is_flat_product():
            for i in itertools.product(*self.set_tuple):
                yield i
        else:
            for i in itertools.product(*self.set_tuple):
                yield pyutilib.misc.flatten_tuple(i)

    def _set_contains(self, element):
        # Do we really need to check if element is a tuple???
        # if type(element) is not tuple:
        #    return False
        try:
            ctr = 0
            for subset in self.set_tuple:
                d = subset.dimen
                if d == 1:
                    if not subset._set_contains(element[ctr]):
                        return False
                else:
                    # cast to tuple is not needed: slices of tuples
                    # return tuples!
                    if not subset._set_contains(element[ctr:ctr+d]):
                        return False

                ctr += d
            return ctr == len(element)
        except:
            return False
        
    def __len__(self):
        ans = 1
        for _set in self.set_tuple:
            ans *= len(_set)
        return ans

    def _compute_dimen(self):
        ans=0
        for _set in self.set_tuple:
            if _set.dimen is None:
                self.dimen=None
                return
            else:
                ans += _set.dimen
        self.dimen = ans

    def is_flat_product(self):
        """
        a simple utility to determine if each of the composite sets is
        of dimension one. knowing this can significantly reduce the
        cost of iteration, as you don't have to call flatten_tuple.
        """

        for s in self.set_tuple:
            if s.dimen != 1:
                return False
        return True

    def _verify(self, element, use_exception=True):
        """
        If this set is virtual, then an additional check is made
        to ensure that the element is in each of the underlying sets.
        """
        tmp = _SetContainer._verify(self, element, use_exception)
        return tmp

        # WEH - when is this needed?
        if not tmp or not self.virtual:
            return tmp

        next_tuple_index = 0
        member_set_index = 0
        for member_set in self.set_tuple:
            tuple_slice = element[next_tuple_index:next_tuple_index + member_set.dimen]
            if member_set.dimen == 1:
                tuple_slice = tuple_slice[0]
            if tuple_slice not in member_set:
                return False
            member_set_index += 1
            next_tuple_index += member_set.dimen
        return True


class _SetArray(Set):
    """An array of sets, which are indexed by other sets"""

    def __init__(self, *args, **kwds):      #pragma:nocover
        """ Constructor """
        self._index_set=None
        self.value=OrderedDict()
        self._ndim=len(args)
        if len(args) == 1:
            if isinstance(args[0],Set):
                self._index=args[0]
            else:
                try:
                    options = getattr(args[0],'set_options')
                    options['initialize'] = args[0]
                    self._index=Set(**options)
                except:
                    self._index=Set(initialize=args[0])
        else:
            self._index=None
            tmp = []
            for arg in args:
                if isinstance(arg,Set):
                    tmp.append(arg)
                else:
                    try:
                        options = getattr(arg,'set_options')
                        options['initialize'] = arg
                        tmp.append( Set(**options) )
                    except:
                        tmp.append( Set(initialize=arg) )
            self._index_set=tuple(tmp)
        Set.__init__(self, **kwds)
        self.value=OrderedDict()

    def keys(self):
        return self.value.keys()

    def __contains__(self, element):
        return element in self.value.keys()

    def clear(self):
        """Remove all elements from the set."""
        for key in self.value:
            self.value[key].clear()

    def data(self):
        """The underlying set data."""
        raise TypeError, "Cannot access underlying set data for array set "+self.name

    def __getitem__(self, key):
        if key not in self.value:
            if (key in self._index) and (self.initialize is not None) and (type(self.initialize) in [tuple,list]):
                return self.initialize
            raise KeyError, "Cannot access index "+str(key)+" in array set "+self.name
        return self.value[key]

    def __setitem__(self, key, val):
        if key not in self._index:
            raise KeyError, "Cannot set index "+str(key)+" in array set "+self.name
        if key in self.value:
            self.value[key].clear()
            for elt in val:
                self.value[key].add(elt)
        else:
            self.value[key] = Set(initialize=val,within=self.domain,validate=self.validate,ordered=self.ordered,name=self.name+"["+str(key)+"]",dimen=self.dimen)
            self.value[key].construct()

    def __len__(self):
        result = 0
        for val in self.value.values():
            result += len(val)
        return result

    def __iter__(self):
        """Return an iterator for the set array"""
        return self.value.__iter__()

    def check_values(self):
        for key in self.value:
            for val in self.value[key]:
                self._verify(val,True)

    def pprint(self, ostream=None, verbose=False):
        if ostream is None:
            ostream = sys.stdout
        print >>ostream, "  ",self.name,":",
        if not self.doc is None:
            print >>ostream, self.doc
            print >>ostream, "  ",
        print >>ostream, "\tDim="+str(self.dim()),
        print >>ostream, "\tDimen="+str(self.dimen),
        if self.domain is None:
            print >>ostream, "\tDomain="+str(self.domain),
        else:
            print >>ostream, "\tDomain="+str(self.domain.name),
        print >>ostream, "\tArraySize="+str(len(self.value.keys())),
        print >>ostream, "\tOrdered="+str(self.ordered)
        if self._model is None:
            print >>ostream, "\t Model=None"
        else:
            print >>ostream, "\t Model="+str(self._model().name)
        tmp=self.value.keys()
        if tmp is not None:
            tmp.sort()
        else:
            tmp=[]
        for val in tmp:
            if self.ordered:
                ttmp = copy.copy(self.value[val].order)
            else:
                ttmp = copy.copy(list(self.value[val].value))
                ttmp.sort()
            print >>ostream, "\t",self.name+"["+str(val)+"]\t",ttmp

    def construct(self, values=None):
        """ Apply the rule to construct values in each set"""
        if __debug__:
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug("Constructing _SetArray, name="+self.name+", from data="+repr(values))
        if self._constructed:
            return
        self._constructed=True
        if self.virtual:                                #pragma:nocover
            raise TypeError, "It doesn't make sense to create a virtual set array"

        if self.initialize is None:
            self.initialize = getattr(self,'rule',None)

        #
        # Construct using the values list
        #
        if values is not None:
            for key in values:
                if type(key) is tuple and len(key)==1:
                    tmpkey=key[0]
                else:
                    tmpkey=key
                if tmpkey not in self._index:
                    raise KeyError, "Cannot set index "+str(tmpkey)+" in array set "+self.name
                self.value[tmpkey] = Set(initialize=values[key],within=self.domain,validate=self.validate,ordered=self.ordered,name=self.name+"["+str(tmpkey)+"]",dimen=self.dimen)
                self.value[tmpkey]._model = self._model
                self.value[tmpkey]._parent = self._parent
                self.value[tmpkey].construct()
        #
        # Construct using the rule
        #
        elif type(self.initialize) is types.FunctionType:
            if self._parent is None:
                raise ValueError, "Need parent block to construct a set array with a function"
            if self._index is None:
                raise ValueError, "No index for set "+self.name
            for key in self._index:
                self.value[key] = Set(within=self.domain,validate=self.validate,ordered=self.ordered,name=self.name+"["+str(key)+"]",dimen=self.dimen)
                self.value[key]._model = self._model
                self.value[key]._parent = self._parent
                self.value[key].construct()
                #
                if isinstance(key,tuple):
                    tmp = key
                else:
                    tmp = (key,)
                if self.initialize.func_code.co_argcount == len(tmp)+1:
                    rule_list = apply_indexed_rule(self, self.initialize, self._parent(), tmp)
                    for val in rule_list:
                        self.value[key].add( val )
                else:
                    ctr=1
                    val = apply_parameterized_indexed_rule(self, self.initialize, self._parent(), ctr, tmp)
                    if val is None:
                        #logger.warning("DEPRECATION WARNING: Set rule returned None instead of Set.Skip")
                        raise ValueError, "Set rule returned None instead of Set.Skip"
                    if not (val.__class__ is tuple and val == Set.End):
                        self.value[key].add( val )
                    while not (val.__class__ is tuple and val == Set.End):
                        self.value[key].add( val )
                        ctr += 1
                        val = apply_parameterized_indexed_rule(self, self.initialize, self._parent(), ctr, tmp)
                        if val is None:
                            #logger.warning("DEPRECATION WARNING: Set rule returned None instead of Set.Skip")
                            raise ValueError, "Set rule returned None instead of Set.Skip"
        #
        # Construct using the default values
        #
        else:
            if self.initialize is not None:
                if type(self.initialize) is not dict:
                    for key in self._index:
                        self.value[key] = Set(initialize=self.initialize,within=self.domain,validate=self.validate,ordered=self.ordered,name=self.name+"["+str(key)+"]",dimen=self.dimen)
                        self.value[key]._model = self._model
                        self.value[key]._parent = self._parent
                        self.value[key].construct()
                else:
                    for key in self.initialize:
                        self.value[key] = Set(initialize=self.initialize[key],within=self.domain,validate=self.validate,ordered=self.ordered,name=self.name+"["+str(key)+"]",dimen=self.dimen)
                        self.value[key]._model = self._model
                        self.value[key]._parent = self._parent
                        self.value[key].construct()






#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the FAST README.txt file.
#  _________________________________________________________________________

from __future__ import division

__all__ = [ 'generate_canonical_repn', 'as_expr', 'canonical_is_constant', 
            'canonical_is_linear', 'canonical_is_quadratic', 'canonical_is_nonlinear' ]

import logging
from coopr.pyomo.base import IPyomoPresolver, IPyomoPresolveAction, Model, \
                             Constraint, Objective, value
from coopr.pyomo.base import param
from coopr.pyomo.base import expr
from coopr.pyomo.base.connector import _ConnectorValue
from coopr.pyomo.base.var import _VarData, _VarElement, Var
import copy


logger = logging.getLogger('coopr.pyomo')
#
# A frozen dictionary that can be hashed.  This dictionary isn't _really_
# frozen, but it acts hashable.
#
class frozendict(dict):

    __slots__ = ['_hash']

    def __hash__(self):
        rval = getattr(self, '_hash', None)
        if rval is None:
            rval = self._hash = hash(frozenset(self.iteritems()))
        return rval

    # getstate and setstate are currently not defined in a generic manner, i.e.,
    # we assume the underlying dictionary is the only thing

    def __getstate__(self):
        return self.items()

    def __setstate__(self, dictionary):
        self.clear()
        self.update(dictionary)
        self._hash = None

#
# Generate a canonical representation of an expression.
#
# The canonical representation is a dictionary.  Each element is a mapping
# from a term degree to terms in the expression.  If the term degree is
# None, then the map value is simply an expression of terms without a
# specific degree.  Otherwise, the map value is a dictionary of terms to
# their coefficients.  A term is represented as a frozen dictionary that
# maps variable id to variable power.  A constant term is represented
# with None.
#
# Examples:
#  Let x[1] ... x[4] be the first 4 variables, and
#      y[1] ... y[4] be the next 4 variables
#
# 1.3                           {0:{ None :1.3}}
# 3.2*x[1]                      {1:{ {0:1} :3.2}}
# 2*x[1]*y[2] + 3*x[2] + 4      {0:{None:4.0}, 1:{{1:1}:3.0}, 2:{{0:1, 5:1}:2.0}}
# 2*x[1]**4*y[2]    + 4         {0:{None:4.0}, 1:{{1:1}:3.0}, 5:{{0:4, 5:1 }:2.0}}
# log(y[1]) + x[1]*x[1]         {2:{{0:2}:1.0}, None:log(y[1])}
#

def generate_canonical_repn(exp, idMap=None):
    degree = exp.polynomial_degree()
    if idMap is None:
        idMap = {}
    if degree == 1:
        # varmap is a map from the variable id() to a _VarData.
        # coef is a map from the variable id() to its coefficient.
        coef, varmap = new_collect_linear_canonical_repn(exp, idMap)
        ans = {}
        if None in coef:
            val = coef.pop(None)
            if val != 0:
                ans[0] = { None: val }
        ans[1] = coef
        ans[-1] = varmap
        return ans
    elif degree > 1:
        ans = collect_general_canonical_repn(exp, idMap)
        if 1 in ans:
            linear_terms = {}
            for key, coef in ans[1].iteritems():
                linear_terms[key.keys()[0]] = coef
            ans[1] = linear_terms
        return ans
    elif degree == 0:
        return { 0: { None:value(exp) } }
    else:
        return { None: exp, -1 : collect_variables(exp, idMap) }


def as_expr(rep, vars=None, model=None, ignore_other=False):
    """ Convert a canonical representation into an expression. """
    if isinstance(model, Model):
        vars = model._var
        id_offset=0
    elif not vars is None:
        id_offset=1
    exp = 0.0
    for d in rep:
        if d is None:
            if not ignore_other:
                exp += rep[d]
            continue
        if d is -1:
            continue
        for v in rep[d]:
            if v is None:
                exp += rep[d][v]
                continue
            e = rep[d][v]
            for id in v:
                for i in xrange(v[id]):
                    if vars is None:
                        e *= rep[-1][id]
                    else:
                        e *= vars[id[1]+id_offset]
            exp += e
    return exp

def repn_add(lhs, rhs, coef=1.0):
    """
    lhs and rhs are the expressions being added together.
    'lhs' and 'rhs' are left-hand and right-hand side operands
    See generate_canonical_repn for explanation of pyomo expressions
    """
    for order in rhs:
        # For each order term, first-order might be 3*x,
        # second order 4*x*y or 5*x**2
        if order is None:
            # i.e., (currently) order is a constant or logarithm
            if order in lhs:
                lhs[order] += rhs[order]
            else:
                lhs[order] = rhs[order]
            continue
        if order < 0:
            # ignore now, handled below
            continue
        if not order in lhs:
            lhs[order] = {}
        for var in rhs[order]:
            # Add coefficients of variables in this order (e.g., third power)
            lhs[order][var] = coef*rhs[order][var] + lhs[order].get(var,0.0)
    #
    # Merge the _VarData maps
    #
    if -1 in rhs:
        if -1 in lhs:
            lhs[-1].update(rhs[-1])
        else:
            lhs[-1] = rhs[-1]
    return lhs

def repn_mult(r1, r2, coef=1.0):
    rep = {}
    for d1 in r1:
        for d2 in r2:
            if d1 == None or d2 == None or d1 < 0 or d2 < 0:
                pass
            else:
                d=d1+d2
                if not d in rep:
                    rep[d] = {}
                if d == 0:
                    rep[d][None] = coef * r1[0][None] * r2[0][None]
                elif d1 == 0:
                    for v2 in r2[d2]:
                        rep[d][v2]  = coef * r1[0][None] * r2[d2][v2] + rep[d].get(v2,0.0)
                elif d2 == 0:
                    for v1 in r1[d1]:
                        rep[d][v1]  = coef * r1[d1][v1] * r2[0][None] + rep[d].get(v1,0.0)
                else:
                    for v1 in r1[d1]:
                        for v2 in r2[d2]:
                            v = frozendict(v1)
                            for id in v2:
                                if id in v:
                                    v[id] += v2[id]
                                else:
                                    v[id]  = v2[id]
                            rep[d][v] = coef * r1[d1][v1] * r2[d2][v2] + rep[d].get(v,0.0)
    #
    # Handle other nonlinear terms
    #
    if None in r1:
        rep[None] = as_expr(r2, ignore_other=True) * copy.deepcopy(r1[None])
        if None in r2:
            rep[None] += copy.deepcopy(r1[None])*copy.deepcopy(r2[None])
    if None in r2:
        if None in rep:
            rep[None] += as_expr(r1, ignore_other=True) * copy.deepcopy(r2[None])
        else:
            rep[None]  = as_expr(r1, ignore_other=True) * copy.deepcopy(r2[None])
    #
    # Merge the _VarData maps
    #
    if -1 in r1:
        rep[-1] = r1[-1]
    if -1 in r2:
        if -1 in rep:
            rep[-1].update(r2[-1])
        else:
            rep[-1] = r2[-1]
    #
    # Return the canonical repn
    #
    return rep

def collect_variables(exp, idMap):
    if exp.is_expression():
        ans = {}
        if exp.__class__ is expr._ProductExpression:
            for subexp in exp._numerator:
                ans.update(collect_variables(subexp, idMap))
            for subexp in exp._denominator:
                ans.update(collect_variables(subexp, idMap))
        else:
            # This is fragile: we assume that all other expression
            # objects "play nice" and just use the _args member.
            for subexp in exp._args:
                ans.update(collect_variables(subexp, idMap))
        return ans
    elif exp.is_fixed():
        # NB: is_fixed() returns True for constants and variables with
        # fixed values
        return {}
    elif exp.__class__ is _VarData or exp.type() is Var:
        key = idMap.setdefault(id(exp), len(idMap))
        return { key : exp }
    else:
        raise ValueError, "Unexpected expression type: "+str(exp)



def _collect_linear_sum(exp, idMap, multiplier, coef, varmap):
    coef[None] += multiplier * exp._const  # None is the constant term in the coefficient map.
    
    arg_coef_iterator = exp._coef.__iter__()
    for arg in exp._args:
        # an arg can be anything - a product, a variable, whatever.

        # Special case... <sigh>
        if arg.__class__ is _VarData and arg.fixed is False:
            # save an expensive recursion - this is by far the most common case.
            _id = id(arg)
            if _id in idMap:
                key = idMap[_id]
            else:
                key = len(idMap)
                idMap[_id] = key
            varmap[key]=arg
            if key in coef:
                coef[key] += multiplier * arg_coef_iterator.next()
            else:
                coef[key] = multiplier * arg_coef_iterator.next()
        else:
            _linear_collectors[arg.__class__](arg, idMap, multiplier * arg_coef_iterator.next(), coef, varmap)

def _collect_linear_prod(exp, idMap, multiplier, coef, varmap):
    multiplier *= exp.coef
    _coef = { None : 0 }
    _varmap = {}
    
    for subexp in exp._denominator:
        x = value(subexp) # only have constants/fixed terms in the denominator.
        if x == 0:
            import StringIO
            buf = StringIO.StringIO()
            subexp.pprint(buf)
            logger.error("Divide-by-zero: offending sub-expression:\n   " + buf)
            raise ZeroDivisionError
        multiplier /= x

    for subexp in exp._numerator:
        if _varmap:
            multiplier *= value(subexp)
        else:
            _linear_collectors[subexp.__class__](subexp, idMap, 1, _coef, _varmap)
            if not _varmap:
                multiplier *= _coef[None]
                _coef[None] = 0

    if _varmap:
        for key, val in _coef.iteritems():
            if key in coef:
                coef[key] += multiplier * val
            else:
                coef[key] = multiplier * val
        varmap.update(_varmap)
    else:
        # constant expression; i.e. 1/x
        coef[None] += multiplier
        

def _collect_linear_pow(exp, idMap, multiplier, coef, varmap):
    if exp.is_fixed():
        coef[None] += multiplier * value(exp)
    elif value(exp._args[1]) == 1:
        arg = exp._args[0]
        _linear_collectors[arg.__class__](arg, idMap, multiplier, coef, varmap)
    else:
        raise TypeError( "Unsupported power expression: "+str(exp._args) )

def _collect_linear_var(exp, idMap, multiplier, coef, varmap):
    if exp.is_fixed():
        coef[None] += multiplier * value(exp)
    else:
        _id = id(exp)
        if _id in idMap:
            key = idMap[_id]
        else:
            key = len(idMap)
            idMap[_id] = key
        if key in coef:
            coef[key] += multiplier
        else:
            coef[key] = multiplier
        varmap[key] = exp

def _collect_linear_const(exp, idMap, multiplier, coef, varmap):
    coef[None] += multiplier * value(exp)

def _collect_linear_connector(exp, idMap, multiplier, coef, varmap):
    pass

def _collect_linear_intrinsic(exp, idMap, multiplier, coef, varmap):
    if exp.is_fixed():
        coef[None] += multiplier * value(exp)
    else:
        raise TypeError( "Unsupported intrinsic expression: %s: %s" % (exp, str(exp._args)) )

_linear_collectors = {
    expr._SumExpression               : _collect_linear_sum,
    expr._ProductExpression           : _collect_linear_prod,
    expr._PowExpression               : _collect_linear_pow,
    expr._IntrinsicFunctionExpression : _collect_linear_intrinsic,
    _ConnectorValue      : _collect_linear_connector,
    param._ParamData     : _collect_linear_const,
    param._ParamElement  : _collect_linear_const,
    param.Param          : _collect_linear_const,
    _VarData             : _collect_linear_var,
    _VarElement          : _collect_linear_var,
    Var                  : _collect_linear_var,
    }

def new_collect_linear_canonical_repn(exp, idMap):
    coef = { None : 0 }
    varmap = {}
    try:
        _linear_collectors[exp.__class__](exp, idMap, 1, coef, varmap)
    except KeyError:
        raise
        raise ValueError( "Unexpected expression (type %s): %s" %
                          (type(exp).__name__, str(exp)) )
    return coef, varmap

def collect_linear_canonical_repn(exp, idMap):
    # we already know this is a linear expression. given the cost of
    # recursion / function calls in Python, do some local
    # optimization. specifically, we know with very high probability
    # (minus some very weird expression trees) that the sub-arguments to
    # a sum expression are either constants or variables.

    # with mutable parameters being the non-default, expressions - and
    # specifically sum expressions - are far more frequent than types
    # such as constants and _VarData. so look at these first, to
    # maximize efficiency.
    if exp.is_expression():
        if exp.__class__ is expr._SumExpression:
            coef = { None : exp._const } # None is the constant term in the coefficient map.
            varmap = {}

            arg_coef_iterator = exp._coef.__iter__()

            for arg in exp._args:

                # an arg can be anything - a product, a variable, whatever.

                # the coefficient should be a float/int. if it was a ParamData object,
                # it would be part of a product expression.
                arg_coef = arg_coef_iterator.next()

                if arg.__class__ is _VarData and arg.fixed is False:
                    # save an expensive recursion - this is by far the most common case.
                    _id = id(arg)
                    if _id in idMap:
                        key = idMap[_id]
                    else:
                        key = len(idMap)
                        idMap[_id] = key
                    varmap[key]=arg
                    if key in coef:
                        coef[key] += arg_coef
                    else:
                        coef[key] = arg_coef
                else:
                    if arg.is_fixed():
                        coef[None] += arg() * arg_coef
                        continue
                    _coef, _varmap = collect_linear_canonical_repn(arg, idMap)
                    varmap.update(_varmap)
                    for var, val in _coef.iteritems():
                        if var in coef:
                            coef[var] += val * arg_coef
                        else:
                            coef[var] = val * arg_coef

            return coef, varmap
        elif exp.__class__ is expr._ProductExpression:
            # OK - the following is safe ONLY because we verify the
            # expression order before entering here -- so we know that
            # there will be no more than 1 variable in the product
            # expression (and that that variable will appear in the
            # numerator!)
            coef = None
            varmap = None
            const = exp.coef
            for subexp in exp._denominator:
                x = value(subexp) # only have constants in the denominator.
                if x == 0:
                    import StringIO
                    buf = StringIO.StringIO()
                    subexp.pprint(buf)
                    logger.error("Divide-by-zero: offending sub-expression:\n   " + buf)
                    raise ZeroDivisionError
                const /= x
            for subexp in exp._numerator:

                if subexp.is_fixed():
                     const *= value(subexp)
                     continue
                elif subexp.__class__ is _VarData:
                    _id = id(subexp)
                    if _id in idMap:
                        key = idMap[_id]
                    else:
                        key = len(idMap)
                        idMap[_id] = key
                    coef, varmap = { key : 1.0 }, { key : subexp }
                else:
                    _coef, _varmap = collect_linear_canonical_repn(subexp, idMap)
                    if _varmap:
                        varmap = _varmap
                        coef = _coef
                    else:
                        const *= _coef[None]

            if coef:
                if const != 1:
                    for var in coef.iterkeys():
                        coef[var] *= const
                return coef, varmap
            else:
                return { None: const }, {}
        elif exp.__class__ is expr._PowExpression:
            # If this is an expression of the form LINEAR_EXP**1,
            # we can just return the representation of LINEAR_EXP
            if exp.is_fixed():
                return { None: value(exp) }, {}
            if value(exp._args[1]) == 1:
                return collect_linear_canonical_repn(exp._args[0], idMap)
            else:
                raise TypeError( "Unsupported power expression: "+str(exp._args) )
        else:
            raise ValueError( "Unsupported expression (type %s):\n%s" %
                              (type(exp).__name__, str(exp)) )
    elif exp.is_fixed():
        # NB: is_fixed() returns True for constants and variables with
        # fixed values
        return { None: value(exp) }, {}
    elif exp.__class__ is _VarData:
        _id = id(exp)
        if _id in idMap:
            key = idMap[_id]
        else:
            key = len(idMap)
            idMap[_id] = key
        return { key : 1.0 }, { key : exp }
    elif exp.type() is Var:
        _id = id(exp)
        if _id in idMap:
            key = idMap[_id]
        else:
            key = len(idMap)
            idMap[_id] = key
        return { key : 1.0 }, { key : exp }
    elif isinstance(exp, _ConnectorValue):
        return {},{}
    else:
        raise ValueError( "Unexpected expression (type %s): %s" %
                          (type(exp).__name__, str(exp)) )


#
# Temporary canonical expressions
#
# temp_const = { 0: {None:0.0} }
# temp_var = { 1: {frozendict({None:1}):1.0} }
# temp_nonl = { None: None }

#
# Internal function for collecting canonical representation, which is
# called recursively.
#
def collect_general_canonical_repn(exp, idMap):
#     global temp_const
#     global temp_var
#     global temp_nonl
    temp_const = { 0: {None:0.0} }
    temp_var = { 1: {frozendict({None:1}):1.0} }
    temp_nonl = { None: None }
    #
    # Ignore identity expressions
    #
    while isinstance(exp, expr._IdentityExpression):
        exp = exp._args[0]
    #
    # Constant
    #
    if exp.is_fixed():
        temp_const[0][None] = value(exp)
        return temp_const
    #
    # Expression
    #
    elif isinstance(exp,expr.Expression):
        #
        # Sum
        #
        if isinstance(exp,expr._SumExpression):
            if exp._const != 0.0:
                repn = { 0: {None:exp._const} }
            else:
                repn = {}
            for i in xrange(len(exp._args)):
                repn = repn_add(repn, collect_general_canonical_repn(exp._args[i], idMap), coef=exp._coef[i] )
            return repn
        #
        # Product
        #
        elif isinstance(exp,expr._ProductExpression):
            #
            # Iterate through the denominator.  If they aren't all constants, then
            # simply return this expresion.
            #
            denom=1.0
            for e in exp._denominator:
                if e.is_fixed():
                    #print 'Z',type(e),e.fixed
                    denom *= e()
                else:
                    temp_nonl[None] = exp
                    return temp_nonl
                if denom == 0.0:
                    print "Divide-by-zero error - offending sub-expression:"
                    e.pprint()
                    raise ZeroDivisionError
            #
            # OK, the denominator is a constant.
            #
            repn = { 0: {None:exp.coef / denom} }
            #print "Y",exp.coef/denom
            for e in exp._numerator:
                repn = repn_mult(repn, collect_general_canonical_repn(e, idMap))
            return repn
        #
        # Power Expression
        #
        elif isinstance(exp,expr._PowExpression):
            if exp.polynomial_degree() is None:
                raise TypeError, "Unsupported general power expression: "+str(exp._args)
                
            # If this is of the form EXPR**1, we can just get the
            # representation of EXPR
            if exp._args[1] == 1:
                return collect_general_canonical_repn(exp._args[0], idMap)
            # The only other way to get a polynomial expression is if
            # exp=EXPR**p where p is fixed a nonnegative integer.  We
            # can expand this expression and generate a canonical
            # representation from there.  If p=0, this expression is
            # constant (and is processed by the is_fixed code above
            # NOTE: There is no check for 0**0
            return collect_general_canonical_repn(
                reduce( lambda x,y: x*y, [exp._args[0]]*int(exp._args[1]),
                        1.0 ), idMap )
        #
        # ERROR
        #
        else:
            raise ValueError, "Unsupported expression type: "+str(exp)
    #
    # Variable
    #
    elif type(exp) is _VarData or exp.type() is Var:
        key = idMap.setdefault(id(exp), len(idMap))
        temp_var = { -1: {key:exp}, 1: {frozendict({key:1}):1.0} }
        return temp_var
    #
    # Connector
    #
    elif type(exp) is _ConnectorValue or exp.type() is Connector:
        # Silently omit constraint...  The ConnectorExpander should
        # expand this constraint into indvidual constraints that
        # reference "real" variables.
        return {}
    #
    # ERROR
    #
    else:
        raise ValueError("Unexpected expression (type %s): " %
                         ( type(exp).__name__, str(exp) ))


def canonical_is_constant(repn):
    """Return True if the canonical representation is a constant expression"""
    return max(repn.iterkeys()) == 0 and None not in repn
    for key in repn:
        if key is None or key > 0:
            return False
    return True

def canonical_is_nonlinear(repn):
    """Return True if the canonical representation is a nonlinear expression"""
    return max(repn.iterkeys()) > 1 or None in repn
    for key in repn:
        if not key in [-1,0,1]:
            return True
    return False

def canonical_is_linear(repn):
    return max(repn.iterkeys()) == 1 and None not in repn
    """Return True if the canonical representation is a linear expression."""
    return (1 in repn) and not is_nonlinear(repn)

def canonical_is_quadratic(repn):
    """Return True if the canonical representation is a quadratic expression."""
    return max(repn.iterkeys()) == 2 and None not in repn
    if not 2 in repn:
        return False
    for key in repn:
        if key is None or key > 2:
            return False
    return True




#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the Coopr README.txt file.
#  _________________________________________________________________________

__all__ = ['SOSConstraint']

import sys
import logging
import weakref

#from expr import *
#from numvalue import *
#from set_types import *
from constraint import _ConstraintData
from indexed_component import IndexedComponent
from misc import create_name, apply_indexed_rule
from pyutilib.component.core import alias


logger = logging.getLogger('coopr.pyomo')



class SOSConstraint(IndexedComponent):
    """
    Represents an SOS-n constraint.

    Usage:
    model.C1 = SOSConstraint(
                             [...],
                             var=VAR,
                             [set=SET OR index=SET],
                             [sos=N OR level=N]
                             )
        [...] Any number of sets used to index SET
        VAR   The set of variables making up the SOS. Indexed by SET.
        SET   The set used to index VAR. SET is optionally indexed by
              the [...] sets. If SET is not specified, VAR is indexed
              over the set(s) it was defined with.
        N     This constraint is an SOS-N constraint. Defaults to 1.

    Example:

      model = AbstractModel()
      model.A = Set()
      model.B = Set(A)
      model.X = Set(B)

      model.C1 = SOSConstraint(model.A, var=model.X, set=model.B, sos=1)

    This constraint actually creates one SOS-1 constraint for each
    element of model.A (e.g., if |A| == N, there are N constraints).
    In each constraint, model.X is indexed by the elements of
    model.D[a], where 'a' is the current index of model.A.

      model = AbstractModel()
      model.A = Set()
      model.X = Var(model.A)

      model.C2 = SOSConstraint(var=model.X, sos=2)

    This produces exactly one SOS-2 constraint using all the variables
    in model.X.
    """


    alias("SOSConstraint", "SOS constraint expressions in a model.")

    def __init__(self, *args, **kwargs):
        name = kwargs.get('name', 'unknown')

        # Get the 'var' parameter
        sosVars = kwargs.pop('var', None)

        # Get the 'set' or 'index' parameters
        if 'set' in kwargs and 'index' in kwargs:
            raise TypeError, "Specify only one of 'set' and 'index' -- " \
                  "they are equivalent parameters"
        sosSet = kwargs.pop('set', None)
        sosSet = kwargs.pop('index', sosSet)

        # Get the 'sos' or 'level' parameters
        if 'sos' in kwargs and 'index' in kwargs:
            raise TypeError, "Specify only one of 'sos' and 'level' -- " \
                  "they are equivalent parameters"
        sosLevel = kwargs.pop('sos', None)
        sosLevel = kwargs.pop('level', sosLevel)

        # Make sure sosLevel has been set
        if sosLevel is None:
            raise TypeError, "SOSConstraint() requires that either the " \
                  "'sos' or 'level' keyword arguments be set to indicate " \
                  "the type of SOS."

        # Make sure we have a variable
        if sosVars is None:
            raise TypeError, "SOSConstraint() requires the 'var' keyword " \
                  "be specified"

        # Find the default sets for sosVars if sosSets is None
        if sosSet is None:
            sosSet = sosVars.index()

        # Construct parents
        kwargs['ctype'] = kwargs.get('ctype', SOSConstraint)
        IndexedComponent.__init__(self, *args, **kwargs)

        # Set member attributes
        self._sosVars = sosVars
        self._sosSet = sosSet
        self._sosLevel = sosLevel

        # TODO figure out why exactly the code expects this to be defined
        # likely due to Numericvalue
        self.domain = None

        # TODO should variables be ordered?

    def construct(self, *args, **kwds):
        """
        A quick hack to call add after data has been loaded. construct
        doesn't actually need to do anything.
        """
        self.add()

    def add(self, *args):
        """
        Mimics Constraint.add, but is only used to alert preprocessors
        to the existence of the SOS variables.

        Ignores all arguments passed to it.
        """

        # self._data is a _ConstraintData object, which has a member
        # .body.  .body needs to have a 3-tuple of expressions
        # containing the variables that are referenced by this
        # constraint. We only use one index, None. I have no
        # particular reason to use None, it just seems consistent with
        # what Constraint objects do.

        # For simplicity, we sum over the variables.

        vars = self.sos_vars()

        expr = sum(vars[i] for i in vars._index)

        self._data[None] = _ConstraintData(create_name(self.name, None), self)
        self._data[None].body = expr

    def sos_level(self):
        """ Return n, where this class is an SOS-n constraint """
        return self._sosLevel

    def sos_vars(self):
        """ Return the variables in the SOS """
        return self._sosVars

    def sos_set(self):
        """ Return the set used to index the variables """
        return self._sosSet

    def sos_set_set(self):
        """ Return the sets used to index the sets indexing the variables """
        return self._index
    
    def reset(self):            #pragma:nocover
        pass                    #pragma:nocover

    def pprint(self, ostream=None, verbose=False):
        if ostream is None:
            ostream = sys.stdout
        print >>ostream, "  ",self.name,":",
        if not self.doc is None:
            print >>ostream, self.doc
            print >>ostream, "  ",
        print >>ostream, "Type="+str(self._sosLevel)
        print >>ostream, "\tVariable: "+self._sosVars.name
        print >>ostream, "\tIndices: ",
        for i in self._sosSet.value:
            print >>ostream, str(i)+" ",
        print >>ostream, ""
